#
#    IBM Confidential
#    OCO Source Materials
#    [5737-H76, 5725-W78, 5900-A1R, 5737-L65]
#    (C) Copyright IBM Corp. 2020  All Rights Reserved.
#    The source code for this program is not published or otherwise divested of its trade secrets,
#    irrespective of what has been deposited with the U.S. Copyright Office.


class GlobalizationUtil:

    @staticmethod
    def get_language():
        language = "en"
        return language
